import numpy as np
from scipy.optimize import minimize

class model(object):
    """
    Define a heteroskedastic linear model and calculate likelihood, gradients, and maximum likelihood estimates of
    parameters.
    :param y: (n) or (n,1) numpy array of phenotype observations
    :param X: (n) or (n,c) numpy array of n observations of variables. This is the design matrix for the fixed mean effects.
    :param V: (n) or (n,v) numpy array of n observations of variables. This is the design matrix for the fixed variance effects.

    :return: heteroskedastic linear model
    """
    def __init__(self,y,X,V):
        # Get sample size
        self.n = X.shape[0]
        # Check shape of arrays
        if V.ndim == 2:
            self.n_fixed_variance = V.shape[1]
        elif V.ndim == 1:
            self.n_fixed_variance = 1
            V = V.reshape((self.n, 1))
        else:
            raise (ValueError('Incorrect dimension of Variance Covariate Array'))
        if X.ndim == 2:
            self.n_fixed_mean = X.shape[1]
        elif X.ndim == 1:
            self.n_fixed_mean = 1
            X = X.reshape((self.n, 1))
        else:
            raise (ValueError('Incorrect dimension of Mean Covariate Array'))
        # phenotype
        self.y=y
        # mean covariates
        self.X=X
        # variance covariates
        self.V=V

    # Compute likelihood of data given beta, alpha
    def likelihood(self,beta,alpha,negative=False):
        """
        Compute the log of the likelihood, the likelihood at the maximum likelihood for the fixed mean effects
        :param alpha: value of fixed mean effects to compute likelihood for
        :param beta: value of fixed variance effects to compute likelihood for
        :param negative: compute -2*L/n-log(2*pi), where L is the log-likelihood, the function that is minimized to find the MLE. Default is False.

        :return: L, log-likelihood of data given parameters.

        """
        Vbeta = self.V.dot(beta)
        resid = self.y - self.X.dot(alpha)
        L = np.sum(Vbeta) + np.sum(np.square(resid) * np.exp(-Vbeta))
        if not negative:
            L = -0.5*(L+self.n*np.log(2*np.pi))
        return L

    # Compute MLE of alpha given beta
    def alpha_mle(self,beta):
        """
                Compute the maximum likelihood estimate of the fixed mean effect parameters, given
                particular fixed variance effect parameters and variance of random effects
                :param beta: value of fixed variance effects


                :return: alpha, 1d numpy array
        """
        D_inv = np.exp(-self.V.dot(beta))
        X_t_D_inv = np.transpose(self.X) * D_inv
        alpha = np.linalg.solve(X_t_D_inv.dot(self.X), X_t_D_inv.dot(self.y))
        return alpha

    # Compute gradient with respect to beta for a given beta and alpha
    def grad_beta(self,beta, alpha):
        """
        Compute the gradient with respect to the fixed variance effects of -2*L/n-log(2*pi),
        where L is the log-likelihood, the function that is minimized to find the MLE
        :param alpha: value of fixed mean effects to compute likelihood for
        :param beta: value of fixed variance effects to compute likelihood for

        :return: grad_beta, 1d numpy array

        """
        D_inv = np.exp(-self.V.dot(beta))
        resid_2 = np.square(self.y - self.X.dot(alpha))
        k = 1 - resid_2 * D_inv
        V_scaled = np.transpose(np.transpose(self.V) * k)
        n1t = np.ones((1, self.X.shape[0]))
        return n1t.dot(V_scaled)

    # OLS solution for alpha
    def alpha_ols(self):
        """
                Compute the ordinary least squares (OLS) estimate of the fixed mean effect parameters

                :return: alpha, 1d numpy array
        """
        # Get initial guess for alpha
        return np.linalg.solve(np.dot(self.X.T, self.X), np.dot(self.X.T, self.y))

    # Find an approximation to the MLE of beta given alpha
    def approx_beta_mle(self):
        """
                Analytical approximation to the maximum likelihood estimate of the fixed variance effects

                :return: beta, 1d numpy array
        """
        # Get alpha OLS
        alpha=self.alpha_ols()
        # squared residuals
        resid_2=np.square(self.y-self.X.dot(alpha)).reshape((self.X.shape[0]))
        # RHS
        V_scaled=np.transpose(np.transpose(self.V)*(resid_2-1))
        n1t=np.ones((1,self.X.shape[0]))
        b=n1t.dot(V_scaled).reshape(self.V.shape[1])
        # LHS
        V_t_scaled=np.transpose(self.V)*resid_2
        A=V_t_scaled.dot(self.V)
        return np.linalg.solve(A,b)

    def beta_cov(self):
        # Find the covariance matrix of beta
        return 2 * np.linalg.inv(np.dot(self.V.T, self.V))

    def alpha_cov(self,beta):
        # Find the covariance matrix for alpha given beta
        D_inv=np.exp(-self.V.dot(beta))
        precision=np.dot(np.transpose(self.X)*D_inv,self.X)
        return np.linalg.inv(precision)

    def optimize_model(self):
        """
                Find the maximum likelihood estimate (MLE) of the parameters and their sampling distribution.

                :return: optim, dictionary with MLEs ('alpha', fixed mean effects; 'beta', fixed variance effects),
                their standard errors ('alpha_se', 'beta_se', 'h2_se'),
                covariance matrix for sampling distribution of parameter vectors ('beta_cov' and 'alpha_cov'),
                maximum likelihood ('likelihood'), whether optimisation was successful ('success'),
        """
        # Get initial guess for beta
        beta_init = self.approx_beta_mle()
        # Optimize
        optimized = minimize(likelihood_beta, beta_init,
                             args=(self.y, self.X, self.V),
                             method='L-BFGS-B',
                             jac=gradient_beta)
        if not optimized.success:
            print('Optimization unsuccessful.')
        # Get MLE
        beta_mle = optimized['x']
        alpha = self.alpha_mle(beta_mle)
        # Get parameter covariance
        optim = {}
        optim['beta'] = optimized['x']
        optim['alpha'] = self.alpha_mle(optim['beta'])
        optim['beta_cov'] = self.beta_cov()
        optim['beta_se'] = np.sqrt(np.diag(optim['beta_cov']))
        optim['alpha_cov'] = self.alpha_cov(beta_mle)
        optim['alpha_se'] = np.sqrt(np.diag(optim['alpha_cov']))
        optim['likelihood'] = -0.5 *self.n * (optimized['fun'] + np.log(2 * np.pi))
        optim['success']=optimized.success
        return optim

##### Functions to pass to opimizer ######
def likelihood_beta(beta,*args):
    # Profile likelihood of hlm_model as a function of beta. To pass to L-BFGS-B
    y,X,V=args
    n=np.float(X.shape[0])
    hlm_mod=model(y,X,V)
    alpha=hlm_mod.alpha_mle(beta)
    return hlm_mod.likelihood(beta,alpha,negative=True)/n


def gradient_beta(beta, *args):
    # Gradient of profile likelihood with respect to beta (at the MLE of alpha). To pass to L-BFGS-B.
    y,X,V=args
    n = np.float(X.shape[0])
    hlm_mod=model(y,X,V)
    alpha = hlm_mod.alpha_mle(beta)
    return hlm_mod.grad_beta(beta, alpha).reshape((hlm_mod.V.shape[1]))/n





"""
Usage: hlmm_chr.py

This script fits heteroskedastic linear models or heteroskedastic linear mixed models to a sequence of genetic variants
contained in a .bed file. You need to specify the genotypes.bed file, which also has genotypes.bim and genotypes.fam in
the same directory, along with the start and end indices of segment you want the script to fit models to.
The script runs from start to end-1 inclusive, and the first SNP has index 0.
The script is designed to run on a chromosome segment to facilitate parallel computing on a cluster.

The phenotype file and covariate file formats are the same: FID, IID, Trait1, Trait2, ...

If you specify a random_gts.bed file with the option --random_gts, the script will model random effects for
all of the variants specified in random_gts.bed. If no --random_gts are specified, then heteroskedastic linear
models are used, without random effects.

Minimally, the script will output a file outprefix.models.gz, which contains a table of the additive
and log-linear variance effects estimated for each variant in the bed file.

If --random_gts are specified, the script will output an estimate of the variance of the random effects
in the null model in outprefix.null_h2.txt. --no_h2_estimate suppresses this output.

If covariates are also specified, it will output estimates of the covariate effects from the null model as
outprefix.null_mean_effects.txt and outprefix.null_variance_effects.txt. --no_covariate_estimates suppresses this output.
"""

from hlmm import hetlm
from hlmm import hetlmm
import argparse
import numpy as np
from pysnptools.snpreader import Bed, Pheno
from scipy.stats import chi2, zscore

####### Output functions ##########
def neglog10pval(x,df):
    return -np.log10(np.e)*chi2.logsf(x,df)

def vector_out(alpha,se,digits=4):
##Output parameter estimates along with standard errors, t-statistics, and -log10(p-values) ##
    ## Calculate test statistics
    # t-statistic
    t=alpha/se
    # chi-square statistic
    x2=np.square(t)
    # Create output strings
    if len(alpha.shape)==0:
        pval=neglog10pval(x2,1)
        alpha_print=str(round(alpha,digits))+'\t'+str(round(se,digits))+'\t'+str(round(t,digits))+'\t'+str(round(pval,digits))
    else:
        pvals=[neglog10pval(x,1) for x in x2]
        alpha_print=''
        for i in xrange(0,len(alpha)-1):
            alpha_print+=str(round(alpha[i],digits))+'\t'+str(round(se[i],digits))+'\t'+str(round(t[i],digits))+'\t'+str(round(pvals[i],digits))+'\t'
        i+=1
        alpha_print+=str(round(alpha[i],digits))+'\t'+str(round(se[i],digits))+'\t'+str(round(t[i],digits))+'\t'+str(round(pvals[i],digits))
    return alpha_print

def id_dict_make(ids):
## Make a dictionary mapping from IDs to indices ##
    if not type(ids)==np.ndarray:
        raise(ValueError('Unsupported ID type: should be numpy nd.array'))
    id_dict={}
    for id_index in xrange(0,len(ids)):
        id_dict[tuple(ids[id_index,:])]=id_index
    return id_dict

def read_covariates(covar_file,ids_to_match,missing):
## Read a covariate file and reorder to match ids_to_match ##
    # Read covariate file
    covar_f = Pheno(covar_file, missing=missing).read()
    # Get covariate values
    n_X=covar_f._col.shape[0]+1
    X=np.ones((covar_f.val.shape[0],n_X))
    X[:, 1:n_X] = covar_f.val
    # Get covariate names
    X_names = np.zeros((n_X), dtype='S10')
    X_names[0] = 'Intercept'
    X_names[1:n_X] = np.array(covar_f._col, dtype='S20')
    # Match with geno_ids
    ids_dict = id_dict_make(np.array(covar_f.iid))
    X_id_match = np.array([ids_dict[tuple(x)] for x in ids_to_match])
    X = X[X_id_match, :]
    return [X,X_names]

######### Command line arguments #########
if __name__ == '__main__':
    parser=argparse.ArgumentParser()
    parser.add_argument('genofile',type=str,help='Path to genotypes in BED format')
    parser.add_argument('start',type=int,help='Index of SNP in genofile from which to start computing test stats')
    parser.add_argument('end',type=int,help='Index of SNP in genofile at which to finish computing test stats')
    parser.add_argument('phenofile',type=str,help='Location of the y file in PLINK format')
    parser.add_argument('outprefix',type=str,help='Location to output csv file with association statistics')
    parser.add_argument('--mean_covar',type=str,help='Location of mean covariate file in PLINK format (default None)',
                        default=None)
    parser.add_argument('--var_covar',type=str,help='Locaiton of variance covariate file in PLINK format (default None)',
                        default=None)
    parser.add_argument('--fit_covariates',action='store_true',
                        help='Fit covariates for each locus. Default is to fit for null model and project out (mean) and rescale (variance)',
                        default=False)
    parser.add_argument('--random_gts',type=str,help='Location of the BED file with the genotypes of the SNPs that random effects should be modelled for',default=None)
    parser.add_argument('--h2_init',type=float,help='Initial value for variance explained by random effects (default 0.05)',
                        default=0.05)
    parser.add_argument('--phen_index',type=int,help='If y file contains multiple ys, which column to choose (default 1, first)',
                        default=1)
    parser.add_argument('--min_maf',type=float,help='Minimum minor allele frequency (default 0.05)',default=0.05)
    parser.add_argument('--missing_char',type=str,help='Missing value string in y file (default NA)',default='NA')
    parser.add_argument('--max_missing',type=float,help='Maximum percent of missing genotype calls (default 5)',default=5)
    parser.add_argument('--append',action='store_true',default=False,help='Append results to existing output file')
    parser.add_argument('--whole_chr',action='store_true',default=False,help='Fit models to all variants in .bed genofile')
    parser.add_argument('--no_covariate_estimates',action='store_true',default=False,help='No output of covariate effect estimates')
    parser.add_argument('--no_h2_estimate',action='store_true',default=False,help='No output of h2 estimate')

    args=parser.parse_args()

    ####################### Read in data #########################
    ### Read genotypes ###
    test_chr=Bed(args.genofile)
    # select subset to test
    if args.whole_chr:
        sid = test_chr.sid
        test_chr = test_chr.read()
    else:
        sid = test_chr.sid[args.start:args.end]
        test_chr = test_chr[:, args.start:args.end].read()
    genotypes=test_chr.val
    # Get genotype matrix
    if genotypes.ndim==1:
        chr_length=1
        genotypes=genotypes.reshape(genotypes.shape[0],1)
    else:
        chr_length=genotypes.shape[1]
    print('Number of test loci: '+str(genotypes.shape[1]))
    # Get sample ids
    geno_ids=test_chr.iid

    #### Read phenotype ###
    pheno=Pheno(args.phenofile,iid_if_none=geno_ids,missing=args.missing_char).read()
    y=np.array(pheno.val)
    if y.ndim==1:
        pass
    elif y.ndim==2:
        y=y[:,args.phen_index-1]
    else:
        raise(ValueError('Incorrect dimensions of phenotype array'))
    print('Number of y observations: '+str(y.shape[0]))
    # Match IDs with geno IDs
    pheno_ids=np.array(pheno.iid)
    pheno_id_dict=id_dict_make(pheno_ids)
    pheno_id_match=np.array([pheno_id_dict[tuple(x)] for x in geno_ids])
    y=y[pheno_id_match]
    # Remove y NAs
    y_not_nan=np.logical_not(np.isnan(y))
    if np.sum(y_not_nan)<y.shape[0]:
        y=y[y_not_nan]
        # Remove NAs from genotypes
        genotypes=genotypes[y_not_nan,:]
    # Get sample size
    n=genotypes.shape[0]
    print(str(n)+' non missing cases from phenotype')
    n=float(n)

    ### Get covariates
    ## Get mean covariates
    if not args.mean_covar==None:
        X, X_names = read_covariates(args.mean_covar,geno_ids,args.missing_char)
        n_X=X.shape[1]
        # Remove rows with missing values
        X=X[y_not_nan,:]
        # Normalise non-constant cols
        X_stds = np.std(X[:,1:n_X],axis=0)
        X[:,1:n_X]=zscore(X[:,1:n_X],axis=0)
    else:
        X=np.ones((int(n),1))
        n_X=1
        X_names=np.array(['Intercept'])
    ## Get variance covariates
    if not args.var_covar==None:
        V, V_names = read_covariates(args.var_covar,geno_ids,args.missing_char)
        n_V=V.shape[1]
        # Remove rows with missing values
        V=V[y_not_nan,:]
        # Normalise non-constant cols
        V_stds = np.std(V[:, 1:n_V], axis=0)
        V[:,1:n_V]=zscore(V[:,1:n_V],axis=0)
    else:
        V=np.ones((int(n),1))
        n_V=1
        V_names=np.array(['Intercept'])
    n_pars=n_X+n_V+1
    print(str(n_pars)+' parameters in model')

    #### Read random effect genotypes ####
    if args.random_gts is not None:
        random_gts_f=Bed(args.random_gts).read()
        G=random_gts_f.val
        # Check for NAs
        random_isnan=np.isnan(G)
        random_gts_NAs=np.sum(random_isnan,axis=0)
        gts_with_obs=list()
        if np.sum(random_gts_NAs)>0:
            print('Mean imputing missing genotypes in random effect design matrix')
            for i in xrange(0,G.shape[1]):
                if random_gts_NAs[i]<G.shape[0]:
                    gts_with_obs.append(i)
                    if random_gts_NAs[i]>0:
                        gt_mean=np.mean(G[np.logical_not(random_isnan[:,i]),i])
                        G[random_isnan[:,i],i]=gt_mean
            # Keep only columns with observations
            G=G[:,gts_with_obs]
        G = zscore(G, axis=0)
        # Rescale random effect design matrix
        G = np.power(G.shape[1], -0.5) * G
        print(str(int(G.shape[1]))+' loci in random effect')
    else:
        G = None

    ######### Initialise output files #######
    ## Output file
    if args.append:
        write_mode='ab'
    else:
        write_mode='wb'
    outfile=open(args.outprefix+'.models.gz',write_mode)
    if not args.append:
        header='SNP\tn\tfrequency\tlikelihood\tadd\tadd_se\tadd_t\tadd_pval\tvar\tvar_se\tvar_t\tvar_pval\tav_pval\n'
        outfile.write(header)

    ######### Fit Null Model ##########
    ## Get initial guesses for null model
    print('Fitting Null Model')
    # Optimize null model
    if G is not None:
        null_optim= hetlmm.model(y, X, V, G).optimize_model(args.h2_init)
    else:
        null_optim= hetlm.model(y, X, V).optimize_model()
    ## Record fitting of null model
    # Get print out for fixed mean effects
    alpha_out=np.zeros((n_X,2))
    alpha_out[:,0]=null_optim['alpha']
    alpha_out[:,1]=null_optim['alpha_se']
    # Rescale
    if n_X>1:
        alpha_out[1:n_X] = alpha_out[1:n_X]/X_stds
    if not args.append and not args.no_covariate_estimates and args.mean_covar is not None:
        np.savetxt(args.outprefix + '.null_mean_effects.txt',
                   np.hstack((X_names.reshape((n_X, 1)), np.array(alpha_out, dtype='S20'))),
                   delimiter='\t', fmt='%s')
    # variance effects
    beta_out=np.zeros((n_V,2))
    beta_out[0:n_V,0]=null_optim['beta']
    beta_out[0:n_V,1]=null_optim['beta_se']
    # Rescale
    if n_V>1:
        beta_out[1:n_X] = beta_out[1:n_X]/V_stds
    if not args.append and not args.no_covariate_estimates and args.var_covar is not None:
        np.savetxt(args.outprefix + '.null_variance_effects.txt',
                   np.hstack((V_names.reshape((n_V, 1)), np.array(beta_out, dtype='S20'))),
                   delimiter='\t', fmt='%s')
    # h2
    if G is not None:
        if not args.append and not args.no_h2_estimate:
            np.savetxt(args.outprefix + '.null_h2.txt',
                       np.array([null_optim['h2'], null_optim['h2_se']], dtype='S20'),
                       delimiter='\t', fmt='%s')

    ### Project out mean covariates and rescale if not fitting for each locus
    if not args.fit_covariates:
        # Residual y
        y=y-X.dot(null_optim['alpha'])
        # Reformulate fixed_effects
        X=np.ones((int(n),1))
        n_X=1
        # Rescaled residual y
        D_null_sqrt=np.exp(0.5*V.dot(null_optim['beta']))
        y=y/D_null_sqrt
        # Reformulate fixed variance effects
        V=np.ones((int(n),1))
        n_V=1

    ############### Loop through loci and fit AV models ######################
    print('Fitting models for specified loci')
    for loc in xrange(0,chr_length):
        # Filler for output if locus doesn't pass thresholds
        additive_av_out='NaN\tNaN\tNaN\tNaN'
        variance_out='NaN\tNaN\tNaN\tNaN'
        likelihood=np.nan
        allele_frq=np.nan
        av_pval=np.nan
        # Get test genotypes
        test_gts=genotypes[:,loc]
        # Find missingness and allele freq
        test_gt_not_na=np.logical_not(np.isnan(test_gts))
        n_l=np.sum(test_gt_not_na)
        missingness = 100.0 * (1 - float(n_l) / n)
        if missingness<args.max_missing:
            test_gts=test_gts[test_gt_not_na]
            test_gts = test_gts.reshape((test_gts.shape[0], 1))
            allele_frq=np.mean(test_gts)/2
            # Mean normalise genotype vector
            test_gts = test_gts - 2*allele_frq
            if allele_frq>0.5:
                allele_frq=1-allele_frq
            if allele_frq>args.min_maf:
                # Remove missing data rows
                y_l=y[test_gt_not_na]
                X_l=X[test_gt_not_na,:]
                V_l=V[test_gt_not_na,:]
                # Add test locus genotypes to mean and variance fixed effect design matrices
                X_l=np.hstack((X_l,test_gts))
                V_l=np.hstack((V_l,test_gts))
                # Record standard deviation of test gt
                print('Fitting AV model for locus '+str(loc))
                if G is not None:
                    G_l = G[test_gt_not_na, :]
                    av_optim = hetlmm.model(y_l, X_l, V_l, G_l).optimize_model(null_optim['h2'])
                else:
                    av_optim= hetlm.model(y_l, X_l, V_l).optimize_model()
                # Check convergence success
                if av_optim['success']:
                    # Likelihood
                    likelihood=av_optim['likelihood']
                    # Mean effect of locus
                    additive_av_out=vector_out(av_optim['alpha'][n_X],av_optim['alpha_se'][n_X],6)
                    # Variance effect of locus
                    variance_out=vector_out(av_optim['beta'][n_V],av_optim['beta_se'][n_V],6)
                    av_pval=neglog10pval((av_optim['alpha'][n_X]/av_optim['alpha_se'][n_X])**2+(av_optim['beta'][n_V]/av_optim['beta_se'][n_V])**2,2)
                else:
                    print('Maximisation of likelihood failed for for '+sid[loc])
        outfile.write(sid[loc] + '\t'+str(n_l)+'\t'+ str(allele_frq)+'\t'+str(likelihood)+'\t'+additive_av_out+'\t'+variance_out+'\t'+str(round(av_pval,6))+'\n')
    outfile.close()


from setuptools import setup

setup(name='hlmm',
      version='1.2.0a1',
      description='Functions for fitting heteroskedastic linear (mixed) models to genetic data',
      url='http://github.com/alexTISYoung/hlmm',
      download_url='https://github.com/AlexTISYoung/hlmm/archive/1.2.0a1.tar.gz',
      author='Alexander I. Young',
      author_email='alextisyoung@gmail.com',
      license='MIT',
      scripts=['bin/hlmm_chr.py'],
      classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Bio-Informatics',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: MIT License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 2.7',
      ],
      keywords='statistics genetics heteroskedastic linear mixed model',
      packages=['hlmm'],
      install_requires=[
            'numpy',
            'scipy',
            'argparse',
            'pysnptools'
        ],
      extras_require={
            'test': ['numdifftools'],
      },
      zip_safe=False)


import numpy as np
from scipy import linalg
from scipy.optimize import fmin_l_bfgs_b
from scipy.stats import zscore

from hlmm import hetlm

"""
Module for defining heteroskedastic linear models, simulating heteroskedastic linear models,
and finding maximum likelihood estimates of the parameters. 

Class: model
"""

class model(object):
    """
    Define a heteroskedastic linear mixed model and calculate likelihood, gradients, and maximum likelihood estimates of
    parameters.
    :param y: (n) or (n,1) numpy array of phenotype observations
    :param X: (n) or (n,c) numpy array of n observations of variables. This is the design matrix for the fixed mean effects.
    :param V: (n) or (n,v) numpy array of n observations of variables. This is the design matrix for the fixed variance effects.
    :param G: (n) or (n,l) numpy array of n observations of variables. This is the design matrix for the random effects.

    :return: heteroskedastic linear mixed model
    """
    def __init__(self,y,X,V,G):
        # Get sample size
        self.n = X.shape[0]
        # Check shape of arrays
        if G.ndim == 2:
            self.l = G.shape[1]
        elif G.ndim == 1:
            self.l = 1
            G = G.reshape((self.n, 1))
        else:
            raise (ValueError('Incorrect dimension of Random Effects Design Matrix'))
        if V.ndim == 2:
            self.n_fixed_variance = V.shape[1]
        elif V.ndim == 1:
            self.n_fixed_variance = 1
            V = V.reshape((self.n, 1))
        else:
            raise (ValueError('Incorrect dimension of Variance Covariate Array'))
        if X.ndim == 2:
            self.n_fixed_mean = X.shape[1]
        elif X.ndim == 1:
            self.n_fixed_mean = 1
            X = X.reshape((self.n, 1))
        else:
            raise (ValueError('Incorrect dimension of Mean Covariate Array'))
        # phenotype
        if y.ndim>1:
            raise(ValueError('Incorrect dimension of phenotype array'))
        self.y = y
        # mean covariates
        self.X = X
        # variance covariates
        self.V = V
        # random effects design matrix
        self.G = G

    def likelihood(self,beta,h2,negative=False):
        """
        Compute the log of the profile likelihood, the likelihood at the maximum likelihood for the fixed mean effects
        :param beta: value of fixed variance effects to compute likelihood for
        :param h2: value of variance explained by random effects to compute likelihood for
        :param negative: compute -2*L/n-log(2*pi), where L is the log-likelihood, the function that is minimized to find the MLE. Default is False.

        :return: L, log-likelihood of data given parameters.

        """

        L = self.likelihood_and_gradient(beta,h2,return_grad=False)
        if not negative:
            L = -0.5*self.n*(L+np.log(2*np.pi))
        return L

    # Compute likelihood of data given beta, alpha
    def likelihood_and_gradient(self,beta,h2,return_grad=True):
        """
                Compute the the function that is minimized to find the MLE, LL=-2*L/n-log(2*pi), where L is the log
                of the profile likelihood, the likelihood at the maximum for the fixed mean effects.
                Further, compute the gradient with respect to the fixed variance effects and the variance of the random effects.
                This forms the basis of the function passed to L-BFGS-B in order to find the maximum likelihood parameter estimates.
                :param beta: value of fixed variance effects
                :param h2: value of variance explained by random effects
                :param return_grad: whether to compute and return the gradient as well as the likelihood function. Default is True.

                :return: [LL,grad], the value of the function to be minimized, LL, and its gradient. The gradient is a 1d numpy array
                that has the gradient with respect to beta first followed by the gradient with respect to h2.
        """
        ## Calculate common variables
        # heteroscedasticity
        Vb = np.dot(self.V, beta)
        D_inv = np.exp(-Vb)
        # Low rank covariance
        G_scaled_T = self.G.T * D_inv
        G_scaled = G_scaled_T.T
        G_cov = np.dot(G_scaled_T, self.G)
        Lambda = np.identity(self.l, float) + h2 * G_cov
        Lambda = linalg.eigh(Lambda, overwrite_a=True, turbo=True)
        logdet_Lambda = np.sum(np.log(Lambda[0]))
        Lambda_inv = inv_from_eig(Lambda)
        ## Calculate MLE of fixed effects
        X_scaled = np.transpose(self.X.T * D_inv)
        alpha = alpha_mle_inner(h2, X_scaled, self.X, self.y, G_scaled, self.G, Lambda_inv)
        ## Residuals
        if self.n_fixed_mean > 1:
            resid = self.y - np.dot(self.X, alpha)
        else:
            yhat = self.X * alpha
            yhat = yhat.reshape(self.y.shape)
            resid = self.y - yhat
        ## Squared residuals
        resid_square = np.square(resid)
        rnd_resid = np.dot(G_scaled_T, resid)
        Lambda_inv_rnd_resid = np.dot(Lambda_inv, rnd_resid)
        ### Calculate likelihood
        L = np.sum(Vb) + np.sum(resid_square * D_inv) + logdet_Lambda - h2 * np.dot(np.transpose(rnd_resid),
                                                                                    Lambda_inv_rnd_resid)
        if return_grad:
            ### Calculate gradient
            grad = np.zeros((self.n_fixed_variance+1))
            # Calculate gradient with respect to beta
            k = var_weight(h2, resid, self.G, Lambda_inv, Lambda_inv_rnd_resid)
            n1t = np.ones((self.n)).reshape((1, self.n))
            grad[0:self.n_fixed_variance] = np.dot(n1t, np.transpose(np.transpose(self.V) * (1 - k * D_inv)))
            # Calculate gradient with respect to h2
            grad[self.n_fixed_variance] = grad_h2_inner(Lambda_inv, G_cov, Lambda_inv_rnd_resid)
            return L/np.float64(self.n), grad/np.float64(self.n)
        else:
            return L/np.float64(self.n)

    # OLS solution for alpha
    def alpha_ols(self):
        """
                Compute the ordinary least squares (OLS) estimate of the fixed mean effect parameters

                :return: alpha, 1d numpy array
        """
        # Get initial guess for alpha
        return np.linalg.solve(np.dot(self.X.T, self.X), np.dot(self.X.T, self.y))

    # Compute MLE of alpha given beta
    def alpha_mle(self,beta,h2):
        """
                Compute the maximum likelihood estimate of the fixed mean effect parameters, given
                particular fixed variance effect parameters and variance of random effects
                :param beta: value of fixed variance effects
                :param h2: value of variance explained by random effects

                :return: alpha, 1d numpy array
        """
        ## Calculate common variables
        # heteroscedasticity
        Vb = np.dot(self.V, beta)
        D_inv = np.exp(-Vb)
        # Low rank covariance
        G_scaled = np.transpose(self.G.T * D_inv)
        G_cov = np.dot(np.transpose(self.G), G_scaled)
        Lambda = np.identity(self.l, float) + h2 * G_cov
        Lambda_inv = linalg.inv(Lambda)
        ## Calculate MLE of fixed effects
        X_scaled = np.transpose(self.X.T * D_inv)
        return alpha_mle_inner(h2, X_scaled, self.X, self.y, G_scaled, self.G, Lambda_inv)



    def optimize_model(self,h2,SEs=True,dx=10**(-6)):
        """
                Find the maximum likelihood estimate (MLE) of the parameters and their sampling distribution.
                :param h2: initial value of variance explained by random effects
                :param SEs: whether to compute sampling distribution of parameter estimates. Default is True.
                :param dx: the step size used to compute the Hessian for the computing the parameter sampling distribution

                :return: optim, dictionary with MLEs ('alpha', fixed mean effects; 'beta', fixed variance effects;
                'h2', variance of random effects), their standard errors ('alpha_se', 'beta_se', 'h2_se'),
                covariance matrix for sampling distribution of parameter vector ('par_cov', in order: alpha, beta, h2),
                maximum likelihood ('likelihood'), whether optimisation was successful ('success'),
                warnings from L-BFGS-B optimisation ('warnflag').
        """
        # Initialise parameters
        init_params=np.zeros((self.n_fixed_variance+1))
        init_params[self.n_fixed_variance]=h2
        # Get initial guess for beta
        init_params[0:self.n_fixed_variance] = hetlm.model(self.y, self.X, self.V).optimize_model()['beta']
        ## Set parameter boundaries
        # boundaries for beta
        parbounds = [(None,None) for i in xrange(0,self.n_fixed_variance)]
        # boundaries for h2
        parbounds.append((0.00001, None))
        # Optimize
        optimized = fmin_l_bfgs_b(func=lik_and_grad_var_pars,x0=init_params,
                                args=(self.y, self.X, self.V, self.G),
                                bounds=parbounds)
        # Get MLE
        optim = {}
        optim['success']=True
        optim['warnflag'] = optimized[2]['warnflag']
        if optim['warnflag']!=0:
            print('Optimization unsuccessful.')
            optim['success']=False
        optim['beta'] = optimized[0][0:self.n_fixed_variance]
        optim['h2'] = optimized[0][self.n_fixed_variance]
        optim['alpha'] = self.alpha_mle(optim['beta'],optim['h2'])
        # Get parameter covariance
        optim['likelihood'] = -0.5*np.float64(self.n)*(optimized[1]+np.log(2*np.pi))

        # Compute parameter covariance
        if SEs:
            optim['par_cov'] = self.parameter_covariance(optim['alpha'],optim['beta'],optim['h2'],dx)
            par_se = np.sqrt(np.diag(optim['par_cov'] ))
            optim['alpha_se'] = par_se[0:self.n_fixed_mean]
            optim['beta_se'] = par_se[self.n_fixed_mean:(self.n_fixed_variance+self.n_fixed_mean)]
            optim['h2_se']=par_se[self.n_fixed_mean+self.n_fixed_variance]

        return optim

    # Calculate covariance and SEs of parameters at the MLE
    def parameter_covariance(self,alpha,beta,h2,dx=10**(-6)):
        ## Compute parameter covariance at maximum likelihood by numerical differentiation of gradients. Called from 'optimize_model'.
        # Calculate intermediate variables
        resid = (self.y - self.X.dot(alpha))
        # Residual Error
        D_inv = np.exp(-self.V.dot(beta))
        G_scaled_T = (self.G.T) * D_inv
        G_cov = G_scaled_T.dot(self.G)
        # Random Effect
        Lambda = np.identity(self.l, float) + h2 * G_scaled_T.dot(self.G)
        Lambda_inv = linalg.inv(Lambda, overwrite_a=True, check_finite=False)
        # Components of alpha gradient calculation
        X_scaled = np.transpose((self.X.T) * D_inv)
        X_grad_alpha = X_scaled - h2 * np.dot(np.dot(G_scaled_T.T, Lambda_inv), G_scaled_T.dot(self.X))
        # Form Hessian matrix
        n_pars=self.n_fixed_mean+self.n_fixed_variance+1
        H = np.zeros((n_pars, n_pars))
        # Calculate alpha components of hessian
        for p in xrange(0, self.n_fixed_mean):
            # Calculate change in alpha gradient
            d = np.identity(self.n_fixed_mean)*dx
            resid_upper = (self.y - self.X.dot(alpha + d[p,:]))
            resid_lower = (self.y - self.X.dot(alpha - d[p,:]))
            H[0:self.n_fixed_mean, p] = (grad_alpha(resid_upper, X_grad_alpha) - grad_alpha(resid_lower, X_grad_alpha)) / (
            2.0 * dx)
            # Calculate change in beta gradient
            H[self.n_fixed_mean:(n_pars - 1), p] = (grad_beta(h2, self.G, G_scaled_T, self.V, D_inv, resid_upper,
                                                         Lambda_inv) - grad_beta(h2, self.G, G_scaled_T, self.V, D_inv,
                                                                                 resid_lower, Lambda_inv)) / (2.0 * dx)
            H[p, self.n_fixed_mean:(n_pars - 1)] = H[self.n_fixed_mean:(self.n_fixed_mean + self.n_fixed_variance), p]
            # Calculate change in h2 gradient
            H[n_pars - 1, p] = (grad_h2_parcov(G_scaled_T, G_cov, resid_upper, Lambda_inv) - grad_h2_parcov(G_scaled_T, G_cov,
                                                                                              resid_lower,
                                                                                              Lambda_inv)) / (2.0 * dx)
            H[p, n_pars - 1] = H[n_pars - 1, p]
        # Calculate beta components of Hessian
        for p in xrange(self.n_fixed_mean, n_pars - 1):
            d = np.identity(self.n_fixed_variance) * dx
            # Changed matrices
            D_inv_upper = np.exp(-self.V.dot(beta + d[p-self.n_fixed_mean,:]))
            D_inv_lower = np.exp(-self.V.dot(beta - d[p-self.n_fixed_mean,:]))
            G_scaled_T_upper = (self.G.T) * D_inv_upper
            G_scaled_T_lower = (self.G.T) * D_inv_lower
            G_cov_upper = np.dot(G_scaled_T_upper, self.G)
            G_cov_lower = np.dot(G_scaled_T_lower, self.G)
            Lambda_inv_upper = linalg.inv(np.identity(self.l, float) + h2 * G_cov_upper, overwrite_a=True,
                                          check_finite=False)
            Lambda_inv_lower = linalg.inv(np.identity(self.l, float) + h2 * G_cov_lower, overwrite_a=True,
                                          check_finite=False)
            # Change in beta gradient
            H[self.n_fixed_mean:(n_pars - 1), p] = (grad_beta(h2, self.G, G_scaled_T_upper, self.V, D_inv_upper, resid,
                                                         Lambda_inv_upper) - grad_beta(h2, self.G, G_scaled_T_lower, self.V,
                                                                                       D_inv_lower, resid,
                                                                                       Lambda_inv_lower)) / (2.0 * dx)
            # Change in h2 gradient
            H[n_pars - 1, p] = (grad_h2_parcov(G_scaled_T_upper, G_cov_upper, resid, Lambda_inv_upper) - grad_h2_parcov(
                G_scaled_T_lower, G_cov_lower, resid, Lambda_inv_lower)) / (2.0 * dx)
            H[p, n_pars - 1] = H[n_pars - 1, p]
        # Calculate h2 components of the Hessian
        Lambda_inv_upper = linalg.inv(np.identity(self.l, float) + (h2 + dx) * G_cov, overwrite_a=True, check_finite=False)
        Lambda_inv_lower = linalg.inv(np.identity(self.l, float) + (h2 - dx) * G_cov, overwrite_a=True, check_finite=False)
        H[n_pars - 1, n_pars - 1] = (grad_h2_parcov(G_scaled_T, G_cov, resid, Lambda_inv_upper) - grad_h2_parcov(G_scaled_T, G_cov,
                                                                                                   resid,
                                                                                                   Lambda_inv_lower)) / (
                                    2.0 * dx)
        par_cov = linalg.inv(0.5 * H, overwrite_a=True, check_finite=False)
        return par_cov



def simulate(n,l,alpha,beta,h2):
    """
            Simulate from a heteroskedastic linear mixed model given a set of parameters. This uses a singular
            value decomposition to do the simulation quickly when l<<n.

            The function simulates fixed and random effects design matrices of specified dimensions with independent Gaussian entries.

            :param n: sample size
            :param l: number of random effects
            :param alpha: value of fixed mean effects
            :param beta: value of fixed variance effects
            :param h2: value of variance explained by random effects

            :return: hetlmm.model class with simulated data and phenotypes for the given parameters.
    """
    if (l>n):
        print('Simulation slow for l>n')
    c = alpha.shape[0]
    v = beta.shape[0]
    X = np.random.randn((n * c)).reshape((n, c))
    V = np.random.randn((n * v)).reshape((n, v))
    G = zscore(np.random.binomial(2,0.5,(n,l)),axis=0)*np.power(l,-0.5)
    G_svd = np.linalg.svd(G,full_matrices=False)
    y = np.sqrt(h2)*G_svd[0].dot(np.random.randn((l))*G_svd[1])+np.random.randn((n))*np.exp(V.dot(beta)/2.0)+X.dot(alpha)
    return model(y,X,V,G)

def lik_and_grad_var_pars(pars,*args):
    # Wrapper for function to pass to L-BFGS-B
    y, X, V, G = args
    hlmm_mod = model(y,X,V,G)
    return hlmm_mod.likelihood_and_gradient(pars[0:hlmm_mod.n_fixed_variance],pars[hlmm_mod.n_fixed_variance])

# Only for positive semi-definite matrix
def inv_from_eig(eig):
    # Compute matrix inverse from eigendecomposition
    U_scaled=eig[1]*np.power(eig[0],-0.5)
    return np.dot(U_scaled,U_scaled.T)


def alpha_mle_inner(h2,X_scaled,X,y,Z_scaled,Z,Lambda_inv):
    # Compute the maximum likelihood estimate of the fixed mean effects from intermediate variables in likelihood_and_gradient
    XtD=np.transpose(X_scaled)
    XtDZ=np.dot(XtD,Z)
    XtDZLambda=np.dot(XtDZ,Lambda_inv)
    X_cov_Z=np.dot(XtDZLambda,np.transpose(XtDZ))
    X_cov=np.dot(XtD,X)
    # Matrix multiplying alpha hat
    A=X_cov-h2*X_cov_Z
    # RHS
    X_cov_y=np.dot(XtD,y)
    Z_cov_y=np.dot(np.transpose(Z_scaled),y)
    X_cov_Z_cov_y=np.dot(XtDZLambda,Z_cov_y)
    b=X_cov_y-h2*X_cov_Z_cov_y
    if len(X.shape)==1:
        alpha=b/A
    else:
        alpha=linalg.solve(A,b)
    return alpha

####### Functions for efficient gradient computation ########

def var_weight(h2,resid,G,Lambda_inv,Lambda_inv_rnd_resid):
    # Compute the the weights needed for computation of the gradient with respect to beta
    cov_diagonal=(G.dot(Lambda_inv) * G).sum(-1)
    # Compute weights coming from inner product in low rank space
    a=G.dot(Lambda_inv_rnd_resid)
    k=np.square(resid)+h2*cov_diagonal+h2*a*(h2*a-2*resid)
    return k

def grad_beta(h2,G,G_scaled_T,V,D_inv,resid,Lambda_inv):
    # Compute gradient with respect to variance of random effects from intermediate variables in parameter_covariance
    n=V.shape[0]
    # Low rank covariance
    rnd_resid=np.dot(G_scaled_T,resid)
    Lambda_inv_rnd_resid=np.dot(Lambda_inv,rnd_resid)
    ### Calculate likelihood
    # Get k variance weights
    k=var_weight(h2,resid,G,Lambda_inv,Lambda_inv_rnd_resid)
    n1t=np.ones((n)).reshape((1,n))
    return np.dot(n1t,np.transpose(np.transpose(V)*(1-k*D_inv)))

def grad_h2_inner(Lambda_inv,Z_cov,Lambda_inv_rnd_resid):
    # Compute gradient with respect to variance of random effects from intermediate variables in likelihood_and_gradient
    dl=0
    # Calculate trace
    dl+=np.sum(Lambda_inv*Z_cov)
    # Calculate inner product
    dl+=-np.sum(np.square(Lambda_inv_rnd_resid))
    return dl

####### Functions for efficient computation of parameter covariance ########
def grad_h2_parcov(G_scaled_T,G_cov,resid,Lambda_inv):
    # Compute gradient with respect to variance of random effects from intermediate variables in parameter_covariance
    rnd_resid=G_scaled_T.dot(resid)
    Lambda_inv_rnd_resid=Lambda_inv.dot(rnd_resid)
    return grad_h2_inner(Lambda_inv,G_cov,Lambda_inv_rnd_resid)


def grad_alpha(resid,X_grad_alpha):
    # Compute gradient with respect to alpha from intermediate variables in parameter_covariance
    return -2*np.dot(resid.T,X_grad_alpha)